@@ -4,6 +4,7 @@ import contextlib
 import dis
 import importlib
 import inspect
+import itertools
 import os
 import random
 import re
@@ -11,8 +12,10 @@ import shutil
 import sys
 import textwrap
 import traceback
-from types import ModuleType
-from typing import Any, Dict, Iterable, List, Optional, Tuple
+from dataclasses import dataclass
+from pathlib import Path
+from types import FunctionType, ModuleType
+from typing import Dict, Iterable, List, Optional, Tuple, Union
 
 import click
 from rich.color import ANSI_COLOR_NAMES
@@ -24,29 +27,41 @@ from rich.syntax import Syntax
 from rich.table import Table
 from rich.text import Text
 
-T_JUMP_COLORS = Dict[int, str]
-JUMP_COLORS = set(
+T_JUMP_COLOR_MAP = Dict[int, str]
+JUMP_COLORS = [
     c for c in ANSI_COLOR_NAMES.keys() if not any(("grey" in c, "black" in c, "white" in c))
-)
+]
 RE_JUMP = re.compile(r"to (\d+)")
 
 INSTRUCTION_GRID_HEADERS = ["OFF", "OPERATION", "ARGS", ""]
 T_INSTRUCTION_ROW = Tuple[Text, ...]
 
 
+NUMBER_COLUMN_WIDTH = 4
+
+
 @click.command()
 @click.argument("target", nargs=-1)
 @click.option(
     "--theme", default="monokai", help="Choose the syntax highlighting theme (any Pygments theme)."
 )
+@click.option(
+    "-p/-P",
+    "--paging/--no-paging",
+    default=None,
+    help="Enable/disable displaying output using the system pager. If not passed explicitly, the pager will automatically be used if the output is taller than your terminal.",
+)
 @click.version_option()
 def cli(
     target: Tuple[str],
     theme: str,
+    paging: Optional[bool],
 ) -> None:
     """
     Display the source and bytecode of the TARGET Python functions.
 
+    If you TARGET a class, its __init__ method will be displayed.
+
     Any number of TARGETs may be passed; they will be displayed sequentially.
     """
     # Make sure the cwd (implicit starting point for the import path) is actually on PYTHONPATH.
@@ -56,26 +71,47 @@ def cli(
 
     console = Console(highlight=True, tab_size=4)
 
-    for idx, disp in enumerate(
-        make_source_and_bytecode_display_for_targets(targets=target, theme=theme)
-    ):
-        console.print(*disp)
+    displays = list(make_source_and_bytecode_displays_for_targets(targets=target, theme=theme))
+    parts = itertools.chain.from_iterable(display.parts for display in displays)
+    total_height = sum(display.height for display in displays)
+
+    if paging is None:  # pragma: no cover
+        paging = total_height > (shutil.get_terminal_size((80, 20)).lines - 5)
+
+    if paging:  # pragma: no cover
+        with console.pager(styles=True):
+            console.print(*parts)
+    else:
+        console.print(*parts)
 
 
-def make_source_and_bytecode_display_for_targets(
+@dataclass(frozen=True)
+class Display:
+    parts: List[Union[Rule, Columns]]
+    height: int
+
+
+def make_source_and_bytecode_displays_for_targets(
     targets: Iterable[str], theme: str
-) -> Iterable[Tuple[Rule, Columns]]:
+) -> Iterable[Display]:
     for func in map(find_function, targets):
         yield make_source_and_bytecode_display(func, theme)
 
 
-def find_function(target: str) -> Any:
+def find_function(target: str) -> FunctionType:
     parts = target.split(".")
 
+    if len(parts) == 1:
+        try:
+            raise target_was_a_module(silent_import(parts[0]))
+        except ModuleNotFoundError as e:
+            # target was not *actually* a module
+            raise click.ClickException(str(e))
+
     # Walk backwards along the split parts and try to do the import.
     # This makes the import go as deep as possible.
     for split_point in range(len(parts) - 1, 0, -1):
-        module_path, object = ".".join(parts[:split_point]), ".".join(parts[split_point:])
+        module_path, target_path = ".".join(parts[:split_point]), ".".join(parts[split_point:])
 
         try:
             obj = silent_import(module_path)
@@ -83,16 +119,16 @@ def find_function(target: str) -> Any:
         except ModuleNotFoundError:
             pass
 
-    for o in object.split("."):
+    for o in target_path.split("."):
         try:
             obj = getattr(obj, o)
-        except AttributeError as e:
+        except AttributeError:
             raise click.ClickException(
                 f"No attribute named {o!r} found on {type(obj).__name__} {obj!r}."
             )
 
     if inspect.ismodule(obj):
-        raise click.ClickException("Cannot disassemble modules. Target a specific function.")
+        raise target_was_a_module(obj)
 
     # If the target is a class, display its __init__ method
     if inspect.isclass(obj):
@@ -115,15 +151,41 @@ def silent_import(module_path: str) -> ModuleType:
             )
 
 
-def make_source_and_bytecode_display(function: Any, theme: str) -> Tuple[Rule, Columns]:
-    bytecode = dis.Bytecode(function)
+def target_was_a_module(module: ModuleType) -> click.ClickException:
+    possibilities = list(find_functions(module))
+
+    if len(possibilities) == 0:
+        return click.ClickException(f"Cannot disassemble modules. Target a specific function.")
+    choice = random.choice(possibilities)
+    suggestion = click.style(f"{choice.__module__}.{choice.__qualname__}", bold=True)
+    return click.ClickException(
+        f"Cannot disassemble modules. Target a specific function, like {suggestion}"
+    )
+
+
+def find_functions(
+    module: ModuleType, top_module: Optional[ModuleType] = None
+) -> Iterable[FunctionType]:
+    for obj in vars(module).values():
+        if (inspect.ismodule(module) and inspect.getmodule(obj) != module) or (
+            inspect.isclass(module) and inspect.getmodule(module) != top_module
+        ):
+            continue
+
+        if inspect.isfunction(obj):
+            yield obj
+        elif inspect.isclass(obj):
+            yield from find_functions(obj, top_module=top_module or module)
+
+
+def make_source_and_bytecode_display(function: FunctionType, theme: str) -> Display:
+    instructions = list(dis.Bytecode(function))
     source_lines, start_line = inspect.getsourcelines(function)
 
-    instructions = list(bytecode)
-    jump_colors = find_jump_colors(instructions)
+    jump_color_map = find_jump_colors(instructions)
 
     code_lines, instruction_rows, line_number_lines = align_source_and_instructions(
-        instructions, jump_colors, source_lines, start_line
+        instructions, jump_color_map, source_lines, start_line
     )
 
     line_numbers = "\n".join(line_number_lines)
@@ -134,18 +196,36 @@ def make_source_and_bytecode_display(function: Any, theme: str) -> Tuple[Rule, C
     bytecode_block = make_bytecode_block(
         instruction_rows,
         block_width=half_width,
-        bgcolor=Syntax.get_theme(theme).get_background_style().bgcolor,
+        bgcolor=Syntax.get_theme(theme).get_background_style().bgcolor.name,
     )
     line_numbers_block = make_nums_block(line_numbers)
 
-    return Rule(title=f"{function.__module__}.{function.__qualname__}"), Columns(
-        renderables=(line_numbers_block, source_block, line_numbers_block, bytecode_block)
+    return Display(
+        parts=[
+            Rule(title=make_title(function, start_line), style=random.choice(JUMP_COLORS)),
+            Columns(
+                renderables=(line_numbers_block, source_block, line_numbers_block, bytecode_block)
+            ),
+        ],
+        height=len(code_lines) + 1,  # the 1 is from the Rule
+    )
+
+
+def make_title(function, start_line: int) -> Text:
+    path = Path(inspect.getmodule(function).__file__)
+    try:
+        path = path.relative_to(Path.cwd())
+    except ValueError:  # path is not under the cwd
+        pass
+
+    return Text.from_markup(
+        f"{type(function).__name__} [bold]{function.__module__}.{function.__qualname__}[/bold] from {path}:{start_line}"
     )
 
 
 def align_source_and_instructions(
     instructions: List[dis.Instruction],
-    jump_colors: T_JUMP_COLORS,
+    jump_color_map: T_JUMP_COLOR_MAP,
     raw_source_lines: List[str],
     start_line: int,
 ) -> Tuple[List[str], List[T_INSTRUCTION_ROW], List[str]]:
@@ -163,7 +243,7 @@ def align_source_and_instructions(
             ]
             nums.extend(
                 (
-                    str(n) if not len(line.strip()) == 0 else ""
+                    str(n).rjust(NUMBER_COLUMN_WIDTH) if not len(line.strip()) == 0 else ""
                     for n, line in enumerate(new_code_lines, start=last_line_idx + 1)
                 )
             )
@@ -175,7 +255,7 @@ def align_source_and_instructions(
             nums.append("")
             source_lines.append("")
 
-        instruction_rows.append(make_instruction_row(instr, jump_colors))
+        instruction_rows.append(make_instruction_row(instr, jump_color_map))
 
     # catch leftover source
     source_lines.extend(raw_source_lines[last_line_idx + 1 - start_line :])
@@ -184,13 +264,13 @@ def align_source_and_instructions(
 
 
 def make_instruction_row(
-    instruction: dis.Instruction, jump_colors: T_JUMP_COLORS
+    instruction: dis.Instruction, jump_color_map: T_JUMP_COLOR_MAP
 ) -> T_INSTRUCTION_ROW:
     return (
-        make_offset(instruction, jump_colors),
+        make_offset(instruction, jump_color_map),
         make_opname(instruction),
-        make_arg(instruction, jump_colors),
-        make_arg_repr(instruction, jump_colors),
+        make_arg(instruction, jump_color_map),
+        make_arg_repr(instruction, jump_color_map),
     )
 
 
@@ -198,7 +278,7 @@ def make_blank_instruction_rows(n: int) -> List[T_INSTRUCTION_ROW]:
     return [(Text(),) * len(INSTRUCTION_GRID_HEADERS)] * n
 
 
-def find_jump_colors(instructions: List[dis.Instruction]) -> T_JUMP_COLORS:
+def find_jump_colors(instructions: List[dis.Instruction]) -> T_JUMP_COLOR_MAP:
     jump_targets = [i.offset for i in instructions if i.is_jump_target]
     jump_colors = {
         j: color for j, color in zip(jump_targets, random.sample(JUMP_COLORS, len(jump_targets)))
@@ -207,14 +287,14 @@ def find_jump_colors(instructions: List[dis.Instruction]) -> T_JUMP_COLORS:
 
 
 def calculate_half_width(line_numbers: str) -> int:
-    full_width = shutil.get_terminal_size().columns - 6
+    full_width = shutil.get_terminal_size().columns - (2 + (NUMBER_COLUMN_WIDTH * 2))
     half_width = (full_width - (max(map(len, line_numbers)) * 2)) // 2
     return half_width
 
 
-def make_offset(instruction: dis.Instruction, jump_colors: T_JUMP_COLORS) -> Text:
+def make_offset(instruction: dis.Instruction, jump_color_map: T_JUMP_COLOR_MAP) -> Text:
     return Text(
-        str(instruction.offset), style=Style(color=jump_colors.get(instruction.offset, None))
+        str(instruction.offset), style=Style(color=jump_color_map.get(instruction.offset, None))
     )
 
 
@@ -222,18 +302,18 @@ def make_opname(instruction: dis.Instruction) -> Text:
     return Text(instruction.opname + "  ")
 
 
-def make_arg(instruction: dis.Instruction, jump_colors: T_JUMP_COLORS) -> Text:
+def make_arg(instruction: dis.Instruction, jump_color_map: T_JUMP_COLOR_MAP) -> Text:
     return Text(
         str(instruction.arg) if instruction.arg is not None else "",
-        style=Style(color=jump_colors.get(instruction.arg)),
+        style=Style(color=jump_color_map.get(instruction.arg)),
     )
 
 
-def make_arg_repr(instruction: dis.Instruction, jump_colors: T_JUMP_COLORS) -> Text:
+def make_arg_repr(instruction: dis.Instruction, jump_color_map: T_JUMP_COLOR_MAP) -> Text:
     match = RE_JUMP.match(instruction.argrepr)
     return Text(
         f"{instruction.argrepr}",
-        style=Style(color=jump_colors.get(int(match.group(1)))) if match else None,
+        style=Style(color=jump_color_map.get(int(match.group(1)))) if match else None,
         no_wrap=True,
     )
 
@@ -286,7 +366,7 @@ def make_bytecode_block(
     return grid
 
 
-def get_own_version() -> str:
+def get_own_version() -> str:  # pragma: versioned
     if sys.version_info < (3, 8):
         import importlib_metadata
     else:
